/*
   Editor functions
   --------------------------------------------------------------------
   VCHE NC - Virtual Console Hex Editor NCURSES version

   Copyright (C) 1998, 1999 Diego Javier Grigna <diego@grigna.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

extern char hex_trans_byte1[];
extern char hex_trans_byte2[];

extern char *hbuffer;

int edit_mask;        /* Edit mask for XOR/etc F* keys                  */
static char *cbuffer; /* Color buffer                                   */
static int tab_flag;  /* Switch from hex to ascii editing and viceversa
                         (i.e. left side = hex, right side = ascii edit */
int ex, ey;           /* X and Y coordinates                            */
static int hx, hy;    /* Hex side x, y                                  */
static int ax, ay;    /* Ascii side x, y                                */
static int f_digit;   /* First or second digit (for hex editing)        */
static long posxyh;   /* xy position on virtual screen (hex side)       */
static long posxya;   /* xy position on virtual screen (ascii side)     */

static void set_edit_fkeys( void);
static void set_all_xy( void);
static void edit_set_colorxy( int what_attr);
static void edit_the_buffer( char c);
static void edit_set_charxy_ascii( char c);
static void edit_set_charxy_hex( char c);
#define edit_set_color_buffer() edit_set_colorxy( cbuffer[ ey * bytes_per_row + ex])
#define edit_set_color_in_buffer( attr) cbuffer[ ey * bytes_per_row + ex] = attr
static void set_editchange_colors( void);
static void edit_set_mask( void);
static void edit_bitops( int bitop);
#define edit_xor() edit_bitops( EDIT_BITOP_XOR)
#define edit_or()  edit_bitops( EDIT_BITOP_OR )
#define edit_and() edit_bitops( EDIT_BITOP_AND)
#define edit_neg() edit_bitops( EDIT_BITOP_NEG)
#define edit_add() edit_bitops( EDIT_BITOP_ADD)
#define edit_sub() edit_bitops( EDIT_BITOP_SUB)
static int edit_ask_save( int pos);
static void edit_save_data( void);

void edit( void)
{
 int c;

 cbuffer = ( char *) allocate_mem( ( LINES - 2 ) * COLS);
 memset( cbuffer, BUF_BBLUE, ( LINES - 2 ) * COLS);

 ll_edit_head = NULL;

 ex = ey  = 0;
 f_digit  = 0;
 tab_flag = 0;

 hex_showfile();
 nc_flush();

 set_edit_fkeys();

 term_unhide_cursor();

 while( 1) {
        if( ex <  0) {
            ex = bytes_per_row - 1;
            ey--;
        }

        if( ex > bytes_per_row - 1) {
            ex = 0;
            ey++;
        }

        edit_set_colorxy( BUF_WHITE);

        c = getch();

        switch( c) {
               case '\t': tab_flag = tab_flag ? 0 : 1;
                          break;
               case KEY_F(2) : if( edit_ask_save( EDIT_ASK_SAVE) == -1) {
                                   edit_set_color_buffer();
                                   term_hide_cursor();
                                   free( cbuffer);
                                   return;
                               }
                               break;
               case KEY_F(3) : edit_set_mask();
                               break;
               case KEY_F(4) : edit_xor(); 
                               break;
               case KEY_F(5) : edit_or();
                               break;
               case KEY_F(6) : edit_and();
                               break;
               case KEY_F(7) : edit_neg();
                               break;
               case KEY_F(8) : edit_add();
                               break;
               case KEY_F(9) : edit_sub();
                               break;
               case KEY_F(10): if( edit_ask_save( EDIT_ASK_QUIT) == -1) {
                                   edit_set_color_buffer();
                                   term_hide_cursor();
                                   free( cbuffer);
                                   return;
                               }
                               break;
               case KEY_UP   : edit_set_color_buffer();
                               ey--;
                               if( ey < 0) {
                                   ey = 0;
                                   file_prev_line();
                                   set_editchange_colors();
                               }
                               break;
               case KEY_DOWN : edit_set_color_buffer();
                               ey++;
                               if( ey > LINES - 3) {
                                   ey = LINES - 3;
                                   file_next_line();
                                   set_editchange_colors();
                               }
                               break;
               case KEY_RIGHT: if( !tab_flag) {
                                   if( f_digit) {
                                       edit_set_color_buffer();
                                       ex++;
                                   }
                                   f_digit = f_digit ? 0 : 1;
                               } else {
                                   edit_set_color_buffer();
                                   ex++;
                               }
                               if( ex > bytes_per_row - 1 && ey == LINES - 3) {
                                   ex = 0;
                                   file_next_line();
                                   set_editchange_colors();
                                   break;
                               }
                               break;
               case KEY_LEFT : if( !tab_flag) {
                                   if( !f_digit) {
                                       edit_set_color_buffer();
                                       ex--;
                                   }
                                   f_digit = f_digit ? 0 : 1;
                               } else {
                                   edit_set_color_buffer();
                                   ex--;
                               }
                               if( ex < 0 && ey == 0) {
                                   ex = bytes_per_row - 1;
                                   file_prev_line();
                                   set_editchange_colors();
                                   break;
                               }
                               break;
               case KEY_HOME : file_go_home();
                               set_editchange_colors();
                               ex = ey = 0;
                               break;
               case KEY_END  : file_go_end();
                               set_editchange_colors();
                               ex = ey = 0;
                               break;
               case KEY_PPAGE: file_prev_block();
                               set_editchange_colors();
                               ex = ey = 0;
                               break;
               case KEY_NPAGE: file_next_block();
                               set_editchange_colors();
                               ex = ey = 0;
                               break;
               default:        edit_the_buffer( c);
                               break;
        }

 } /* end while( 1) */

}

static void set_edit_fkeys( void)
{
 int sp, x, y;

 sp = ( COLS / 10) - 8;
 y = LINES - 1;

 nc_color_puts( 0, LINES - 1, COLS, replicate( ' ', COLS), NC_COLOR_BACKCYAN);

 x  = 2;
 x += 8 + sp; nc_color_puts( x, y, 4, "SAVE"   , NC_COLOR_BACKCYAN); /* F2  */
 x += 8 + sp; nc_color_puts( x, y, 6, "EDMASK" , NC_COLOR_BACKCYAN); /* F3  */
 x += 8 + sp; nc_color_puts( x, y, 3, "XOR"    , NC_COLOR_BACKCYAN); /* F4  */
 x += 8 + sp; nc_color_puts( x, y, 2, "OR"     , NC_COLOR_BACKCYAN); /* F5  */
 x += 8 + sp; nc_color_puts( x, y, 3, "AND"    , NC_COLOR_BACKCYAN); /* F6  */
 x += 8 + sp; nc_color_puts( x, y, 3, "NEG"    , NC_COLOR_BACKCYAN); /* F7  */
 x += 8 + sp; nc_color_puts( x, y, 3, "ADD"    , NC_COLOR_BACKCYAN); /* F8  */
 x += 8 + sp; nc_color_puts( x, y, 3, "SUB"    , NC_COLOR_BACKCYAN); /* F9  */
 x += 8 + sp; nc_color_puts( x, y, 6, "CANCEL" , NC_COLOR_BACKCYAN); /* F10 */

 x  = 0;      nc_color_puts( x, y, 2, " 1", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, " 2", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, " 3", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, " 4", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, " 5", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, " 6", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, " 7", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, " 8", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, " 9", NC_LIGHTGRAY );
 x += 8 + sp; nc_color_puts( x, y, 2, "10", NC_LIGHTGRAY );
}

static void set_all_xy( void)
{
 /* The hex numbers begin at ey + 1 */
 hy = ey + 1;
 hx = 3.25 * ex + 12;

 /*
  *  Q: Where the f*** you get "hx = 3.25 * ex + 12" ???????????????
  *  A: Simple!
  *
  *  01234567890123456789012345678901234567890123456789012345678901
  *
  *  0000000000  68 65 6C 6C  6F 20 77 6F  72 6C 64 20  21 21 21 21  hello world
  *  0000000010  21 21 21 21  21 21 0A 0A  0A 0A       |             !!!!!!
  *              |              |__|__|   |____________|
  *              |____the hex      |                   |
  *                   numbers      |                   |
  *                   begin at     |                   |
  *                   ex = 12      |               at ex / 4 we
  *                                |               add one space
  *                             they are
  *                             separated
  *                             by ex * 2
  *
  *  Then we have this formula:   hx = ex + 12 + ( ex * 2) + ( ex / 4);
  *  We could render it this way: hx = ex + (2 * ex) + (1/4 * ex) + 12
  *  We simplify:                 hx = (1 + 2 + 1/4) * ex + 12
  *  Solution:                    hx = 3.25 * ex + 12
  *
  */

 ay = hy;
 ax = ex + COLS - bytes_per_row;

 posxyh = 4 + 2 *( hy * COLS + hx);
 posxya = 4 + 2 *( ay * COLS + ax);
}

/*
 * Position the cursor and set the color of the hex
 * numbers
 */
static void edit_set_colorxy( int what_attr)
{
 attr_t eattr = 0;
 short ecolor = 0;
 chtype c = 0;

 set_all_xy();

 translate_buffer_colors( what_attr, &eattr, &ecolor);

 term_goto( hx, hy);
 c = inch();
 c &= A_CHARTEXT;
 attrset( COLOR_PAIR( ecolor) | eattr);
 addch( c);
 
 term_goto( hx + 1, hy);
 c = inch();
 c &= A_CHARTEXT;
 attrset( COLOR_PAIR( ecolor) | eattr);
 addch( c);

 term_goto( ax, ay);
 c = inch();
 c &= A_CHARTEXT;
 attrset( COLOR_PAIR( ecolor) | eattr);
 addch( c);

 refresh();

 if( tab_flag)
     term_goto( ax, ay);
 else {
     if( f_digit) 
         term_goto( hx + 1, hy);
     else
         term_goto( hx, hy);
 }
}

void set_char_colorxy( char byte, int what_attr)
{
 int posh, posa;

 set_all_xy();

 posh = posxyh - (4 + 2 * COLS);
 posa = posxya - (4 + 2 * COLS);

 obuffer[ posh    ] = hex_trans_byte1[ (unsigned char) byte];
 obuffer[ posh + 1] = what_attr;
 obuffer[ posh + 2] = hex_trans_byte2[ (unsigned char) byte];
 obuffer[ posh + 3] = what_attr;

 obuffer[ posa    ] = byte > 31 && byte < 127 ? byte : '.';
 obuffer[ posa + 1] = what_attr;
}

static void edit_the_buffer( char c)
{

 if( ( ey * bytes_per_row + ex) >= bread) {
     term_beep();
     return;
 }

 if( tab_flag) { /* ascii edit */
     edit_set_charxy_ascii( c);
 } else {        /* hex edit   */
     if( isxdigit( c))
         edit_set_charxy_hex( c);
     else
         term_beep();
 }

}

static void edit_set_charxy_ascii( char c)
{
 struct ll_edit *node = ll_edit_head;
 struct ll_edit *tmp  = ll_edit_head;
 int pos;
 int temp;

 set_all_xy();

 pos = ey * bytes_per_row + ex;

 while( node != NULL) {
        if( node->filepos < ( fpos + pos)) {
            tmp  = node;
            node = node->next;
        } else
            break;
 }

 temp = (unsigned char) c;

 attrset( NC_WHITE);
 mvaddch( hy, hx    , hex_trans_byte1[ temp]);
 mvaddch( hy, hx + 1, hex_trans_byte2[ temp]);

 mvaddch( ay, ax, c > 31 && c < 127 ? c : '.');

 refresh();

 if( hbuffer[ pos] != c) {

     if( node == NULL)
         ll_edit_add( c, fpos + pos, tmp);
     else 
         if( node->filepos == ( fpos + pos)){
             node->byte = c;
         }
         else
             ll_edit_add( c, fpos + pos, tmp);

     edit_set_colorxy( BUF_BLGRAY);
     edit_set_color_in_buffer( BUF_BLGRAY);
 } else {
     if( node != NULL) {
         if( node->filepos == ( fpos + pos)) {
             tmp->next = node->next;
             if( ll_edit_head == node)
                 ll_edit_head = tmp->next;
             free( node);
         }
     }
     edit_set_colorxy( BUF_BBLUE);
     edit_set_color_in_buffer( BUF_BBLUE);
 }

 ex++;
 if( ex > bytes_per_row - 1 && ey == LINES - 3) {
     ex = 0;
     file_next_line();
     set_editchange_colors();
 }

}

static void edit_set_charxy_hex( char c)
{
 struct ll_edit *node = ll_edit_head;
 struct ll_edit *tmp  = ll_edit_head;
 char buf[ 16];
 int temp;
 int pos;
 int b1, b2;

 set_all_xy();

 pos = ey * bytes_per_row + ex;

 while( node != NULL) {
        if( node->filepos < ( fpos + pos)) {
            tmp  = node;
            node = node->next;
        } else
            break;
 }

 if( node == NULL) {
     buf[ 0] = hex_trans_byte1[ (unsigned char) hbuffer[ pos]];
     buf[ 2] = hex_trans_byte2[ (unsigned char) hbuffer[ pos]];
 } else {
     if( node->filepos == ( fpos + pos)) {
         buf[ 0] = hex_trans_byte1[ (unsigned char) node->byte];
         buf[ 2] = hex_trans_byte2[ (unsigned char) node->byte];
     } else {
         buf[ 0] = hex_trans_byte1[ (unsigned char) hbuffer[ pos]];
         buf[ 2] = hex_trans_byte2[ (unsigned char) hbuffer[ pos]];
     }
 }

 c = toupper( c);

 if( f_digit)
     buf[ 2] = c;
 else
     buf[ 0] = c;

 b1 = ((buf[ 0] >= '0' && buf[ 0] <= '9') ? buf[ 0] - '0' : ((buf[ 0] >= 'A' && buf[ 0] <= 'F') ? buf[ 0] - '7': 0));
 b2 = ((buf[ 2] >= '0' && buf[ 2] <= '9') ? buf[ 2] - '0' : ((buf[ 2] >= 'A' && buf[ 2] <= 'F') ? buf[ 2] - '7': 0));

 temp = b1 * 16 + b2;

 attrset( NC_WHITE);
 mvaddch( hy, hx    , buf[ 0]);
 mvaddch( hy, hx + 1, buf[ 2]);

 mvaddch( ay, ax, temp > 31 && temp < 127 ? temp : '.');

 refresh();

 if( hbuffer[ pos] != (char) temp) {
     if( node == NULL)
         ll_edit_add( temp, fpos + pos, tmp);
     else
         if( node->filepos == ( fpos + pos))
             node->byte = temp;
         else
             ll_edit_add( temp, fpos + pos, tmp);

     edit_set_colorxy( BUF_BLGRAY);
     edit_set_color_in_buffer( BUF_BLGRAY);
 } else {
     if( node != NULL) {
         if( node->filepos == ( fpos + pos)) {
             tmp->next = node->next;
             if( ll_edit_head == node)
                 ll_edit_head = tmp->next;
             free( node);
         }
     }
     edit_set_colorxy( BUF_BBLUE);
     edit_set_color_in_buffer( BUF_BBLUE);
 }

 if( f_digit) {
     ex++;
     if( ex > bytes_per_row - 1 && ey == LINES - 3) {
         ex = 0;
         file_next_line();
         set_editchange_colors();
     }
 }

 f_digit = f_digit ? 0 : 1;
}

static void set_editchange_colors( void)
{
 struct ll_edit *node = ll_edit_head;
 int tmpx, tmpy;
 int bytes;

 memset( cbuffer, BUF_BBLUE, ( LINES - 2 ) * COLS);

 while( node != NULL) {
        if( node->filepos < fpos)
            node = node->next;
        else
            break;
 }

 if( node != NULL) {
     tmpx = ex;
     tmpy = ey;

     bytes = node->filepos - fpos;

     while( bytes < bread && node != NULL && node->filepos <= (fpos + bytes)) {

          ey = bytes / bytes_per_row;
          ex = - ey * bytes_per_row + bytes;

          set_char_colorxy( node->byte, BUF_BLGRAY);
          edit_set_color_in_buffer( BUF_BLGRAY);
          node = node->next;
          ex++;
          if( ex > bytes_per_row - 1) {
              ex = 0;
              ey++;
          }
          if( node != NULL)
              bytes = node->filepos - fpos;
     }

     ex = tmpx;
     ey = tmpy;
 }

 nc_flush();
}

static void edit_set_mask( void)
{
 WINDOW *w;
 char emask[ 256];

 bzero( &emask, sizeof( emask));

 sprintf( emask, "0x%X", edit_mask);

 w = nc_dialog_box( "Enter edit mask: ", emask, 38, NC_COLOR_BACKCYAN, NC_WHITE);

 term_wgoto( w, 2, 2);
 key_scanf( w, emask, 38);
 term_unhide_cursor();

 if( !strncasecmp( emask, "0x", 2)) {
     sscanf( emask, "%x", &edit_mask);
 }
 else
     edit_mask = atoi( emask);

 delwin( w);
}

static void edit_bitops( int bitop)
{
 struct ll_edit *node = ll_edit_head;
 struct ll_edit *tmp  = ll_edit_head;
 char buf[ 16];
 char c = 0;
 int pos;
 int temp;

 set_all_xy();

 pos = ey * bytes_per_row + ex;

 if( pos >= bread) {
     term_beep();
     return;
 }

 while( node != NULL) {
        if( node->filepos < ( fpos + pos)) {
            tmp  = node;
            node = node->next;
        } else
            break;
 }

 if( node != NULL && node->filepos == ( fpos + pos))
     switch( bitop) {
             case EDIT_BITOP_XOR: c =  node->byte ^ edit_mask;
                                  break;
             case EDIT_BITOP_OR : c =  node->byte | edit_mask;
                                  break;
             case EDIT_BITOP_AND: c =  node->byte & edit_mask;
                                  break;
             case EDIT_BITOP_NEG: c = ~node->byte;
                                  break;
             case EDIT_BITOP_ADD: c = node->byte;
                                  c++; 
                                  break;
             case EDIT_BITOP_SUB: c = node->byte;
                                  c--;
                                  break;
     }
 else
     switch( bitop) {
             case EDIT_BITOP_XOR: c =  hbuffer[ pos] ^ edit_mask;
                                  break;
             case EDIT_BITOP_OR : c =  hbuffer[ pos] | edit_mask;
                                  break;
             case EDIT_BITOP_AND: c =  hbuffer[ pos] & edit_mask;
                                  break;
             case EDIT_BITOP_NEG: c = ~hbuffer[ pos];
                                  break;
             case EDIT_BITOP_ADD: c = hbuffer[ pos];
                                  c++;
                                  break;
             case EDIT_BITOP_SUB: c = hbuffer[ pos];
                                  c--;
                                  break;
     }

 temp = (unsigned char) c;

 buf[ 0] = hex_trans_byte1[ temp];
 buf[ 2] = hex_trans_byte2[ temp];

 attrset( NC_WHITE);
 mvaddch( hy, hx    , buf[ 0]);
 mvaddch( hy, hx + 1, buf[ 2]);

 mvaddch( ay, ax, c > 31 && c < 127 ? c : '.');

 refresh();

 if( hbuffer[ pos] != c) {

     if( node == NULL)
         ll_edit_add( c, fpos + pos, tmp);
     else
         if( node->filepos == ( fpos + pos)){
             node->byte = c;
         }
         else
             ll_edit_add( c, fpos + pos, tmp);

     edit_set_colorxy( BUF_BLGRAY);
     edit_set_color_in_buffer( BUF_BLGRAY);
 } else {
     if( node != NULL) {
         if( node->filepos == ( fpos + pos)) {
             tmp->next = node->next;
             if( ll_edit_head == node)
                 ll_edit_head = tmp->next;
             free( node);
         }
     }
     edit_set_colorxy( BUF_BBLUE);
     edit_set_color_in_buffer( BUF_BBLUE);
 }

 ex++;
 if( ex > bytes_per_row - 1 && ey == LINES - 3) {
     ex = 0;
     file_next_line();
     set_editchange_colors();
 }

}

static int edit_ask_save( int pos)
{
 int i;

 if( ll_edit_head == NULL)
     return -1;

 i = nc_ask( "The file was modified, save the changes?", " Yes ", " No ", " Cancel ", pos);

 if( i == 0) {
     edit_save_data();
     return -1;
 } else 
     if( i == 1)
         return -1;

 return 0;
}

static void edit_save_data( void)
{
 WINDOW *w;
 struct ll_edit *node = ll_edit_head;
 struct ll_edit *temp = ll_edit_head;

 w = nc_show_message( " Saving... ", 0);

 while( node != NULL) {
        lseek( fd, node->filepos, SEEK_SET);
        write( fd, &node->byte, 1);
        temp = node;
        node = node->next;
        free( temp);
 }

 lseek( fd, (long) fpos, SEEK_SET);

 ll_edit_head = NULL;

 delwin( w);
}

