/*
   Help functions
   --------------------------------------------------------------------
   VCHE NC - Virtual Console Hex Editor NCURSES version

   Copyright (C) 1998, 1999 Diego Javier Grigna <diego@grigna.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

static char *meaning[] = { 
 " NUL "," SOH "," STX "," ETX "," EOT "," ENQ "," ACK ",
 " BEL "," BS  "," HT  "," LF  "," VT  "," FF  "," CR  ",
 " SO  "," SI  "," DLE "," DC1 "," DC2 "," DC3 "," DC4 ",
 " NAK "," SYN "," ETB "," CAN "," EM  "," SUB "," ESC ",
 " FS  "," GS  "," RS  "," US  ","SPACE",
};

/*
 * Function prototypes
 */
static char *int_to_escape( int c);
static char *int_to_binary( int c);

/*
 * VCHE main help
 */
void help( void)
{
 char hlptxt[ 2048];

 sprintf( hlptxt, "\n  VCHE NC - Virtual Console Hex Editor NCURSES version\n\
\n\
  Release %s\n\
\n\
  If you need help, type \"man vche-nc\" at the shell prompt.\n\n\
  It is the NCURSES version of VCHE, to see all the 256 characters\n\
  run \"vche\" instead.\n\n\
  If you have found a bug, have problems or comments, mail\n\
  me to <diego@grigna.com>\n", VCHE_VERSION);

 nc_show_screentext( hlptxt, " VCHE-NC HELP ");
 refresh();
}

void show_gnu_license( void)
{
 char *gnugpl = "\n  VCHE NC - Virtual Console Hex Editor NCURSES version\n\
  Copyright (C) 1998, 1999 Diego Javier Grigna <diego@grigna.com>\n\
\n\
  This program is free software; you can redistribute it and/or modify\n\
  it under the terms of the GNU General Public License as published by\n\
  the Free Software Foundation; either version 2 of the License, or\n\
  (at your option) any later version.\n\
\n\
  This program is distributed in the hope that it will be useful,\n\
  but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
  GNU General Public License for more details.\n\
\n\
  You should have received a copy of the GNU General Public License\n\
  along with this program; if not, write to the Free Software\n\
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n";

 nc_show_screentext( gnugpl, " GNU General Public License ");
 refresh();
}

/*
 * Shows the ascii table and let people navigate it.
 */
void ascii_table( void)
{
 WINDOW *w; 
 WINDOW *w2; 
 char buf[ 128];
 char *bottom_line = "Use         to move           ENTER or F10 to QUIT";
 int slines = LINES - 9; /* Number of lines to show */
 int endpos;
 int pos = 0;
 int ll;
 int i, j;
 int c;

 endpos = 256 - slines;
 if( endpos < 0) endpos = 0;

 w2 = nc_box( 0, 0, COLS, LINES, NC_COLOR_BACKCYAN);

 nc_put_bottom_line( w2, bottom_line);

 wattrset( w2, NC_COLOR_BACKGRAY);

 ll = ( COLS - strlen( bottom_line)) / 2;
 mvwaddch( w2, LINES - 1, ll + 24, ACS_VLINE);
 mvwaddch( w2, LINES - 1, ll +  4, ACS_UARROW);
 mvwaddch( w2, LINES - 1, ll +  6, ACS_DARROW);
 mvwaddch( w2, LINES - 1, ll +  8, ACS_RARROW);
 mvwaddch( w2, LINES - 1, ll + 10, ACS_LARROW);

 wrefresh( w2);
 
 w = nc_box( 0, 0, COLS, LINES - 1, NC_COLOR_BACKCYAN);
 nc_do_title( w, " ASCII Table ");

 ll = LINES - 4;

 wattrset( w, NC_COLOR_BACKCYAN);
 
 mvwaddnstr( w, 3, 9, "Dec   Oct   Hex   Character   Meaning   C Escape     Binary", 59);

 mvwaddch( w,  2,  7, ACS_ULCORNER);
 mvwaddch( w,  2, 71, ACS_URCORNER);
 mvwaddch( w,  4,  7,     ACS_LTEE);
 mvwaddch( w,  4, 71,     ACS_RTEE);
 mvwaddch( w, ll,  7, ACS_LLCORNER);
 mvwaddch( w, ll, 71, ACS_LRCORNER);

 for( i = 1; i < 64; i++) {
      mvwaddch( w,  2, 7 + i, ACS_HLINE);
      mvwaddch( w,  4, 7 + i, ACS_HLINE);
      mvwaddch( w, ll, 7 + i, ACS_HLINE);
 }

 for( i = 5; i < LINES - 4; i++) {
      mvwaddch( w, i,  7, ACS_VLINE);
      mvwaddch( w, i, 71, ACS_VLINE);
 }

 mvwaddch( w, 2, 13, ACS_TTEE);
 mvwaddch( w, 2, 19, ACS_TTEE);
 mvwaddch( w, 2, 25, ACS_TTEE);
 mvwaddch( w, 2, 37, ACS_TTEE);
 mvwaddch( w, 2, 47, ACS_TTEE);
 mvwaddch( w, 2, 58, ACS_TTEE);

 mvwaddch( w, 3,  7, ACS_VLINE);
 mvwaddch( w, 3, 13, ACS_VLINE);
 mvwaddch( w, 3, 19, ACS_VLINE);
 mvwaddch( w, 3, 25, ACS_VLINE);
 mvwaddch( w, 3, 37, ACS_VLINE);
 mvwaddch( w, 3, 47, ACS_VLINE);
 mvwaddch( w, 3, 58, ACS_VLINE);
 mvwaddch( w, 3, 71, ACS_VLINE);

 mvwaddch( w, 4, 13, ACS_PLUS);
 mvwaddch( w, 4, 19, ACS_PLUS);
 mvwaddch( w, 4, 25, ACS_PLUS);
 mvwaddch( w, 4, 37, ACS_PLUS);
 mvwaddch( w, 4, 47, ACS_PLUS);
 mvwaddch( w, 4, 58, ACS_PLUS);

 mvwaddch( w, ll, 13, ACS_BTEE);
 mvwaddch( w, ll, 19, ACS_BTEE);
 mvwaddch( w, ll, 25, ACS_BTEE);
 mvwaddch( w, ll, 37, ACS_BTEE);
 mvwaddch( w, ll, 47, ACS_BTEE);
 mvwaddch( w, ll, 58, ACS_BTEE);

 wrefresh( w);

 j = 5;

 while( 1) {
        if( pos > endpos) {
            term_beep();
            pos = endpos;
        }

        if( pos < 0) {
            term_beep();
            pos = 0;
        }

        if( pos == 0) {
            nc_color_wputc( w, 10,         4, ACS_HLINE , NC_COLOR_BACKCYAN );
            nc_color_wputc( w, 10, LINES - 4, ACS_DARROW, NC_COLOR_ATTENTION);
        } else 
            if( pos == endpos) {
                nc_color_wputc( w, 10,         4, ACS_UARROW, NC_COLOR_ATTENTION);
                nc_color_wputc( w, 10, LINES - 4, ACS_HLINE , NC_COLOR_BACKCYAN );
            } else {
                nc_color_wputc( w, 10,         4, ACS_UARROW, NC_COLOR_ATTENTION);
                nc_color_wputc( w, 10, LINES - 4, ACS_DARROW, NC_COLOR_ATTENTION);
            }

        for( i = pos; i < pos + slines; i++) {
             if( j > LINES - 5)
                 j = 5;

             if( i < 256) {
                 sprintf( buf, "%3d   %03o   %02X        %c        %s      %s      %s"
                           , i, i, i, ( i > 31) && ( i < 127) ? i : '.'
                           , (i == 127) ? " DEL ": i <= 32 ? meaning[ i]: "     "
                           , int_to_escape( i)
                           , int_to_binary( i));
                 wattrset( w, NC_COLOR_BACKCYAN);
                 mvwaddnstr( w, j, 9, buf, 60);
             }

             mvwaddch( w, j, 13, ACS_VLINE);
             mvwaddch( w, j, 19, ACS_VLINE);
             mvwaddch( w, j, 25, ACS_VLINE);
             mvwaddch( w, j, 37, ACS_VLINE);
             mvwaddch( w, j, 47, ACS_VLINE);
             mvwaddch( w, j, 58, ACS_VLINE);
             j++;
        }
        wrefresh( w);

        c = getch();

        switch( c) {
                case '\n':
                case  'q':
                case  'Q':
                case KEY_F(10): delwin( w);
                                delwin( w2);
                                refresh();
                                return;
                                break;
                case KEY_UP   :
                case KEY_LEFT : pos--;
                                break;
                case KEY_DOWN :
                case KEY_RIGHT: pos++;
                                break;
                case KEY_PPAGE: pos -= slines;
                                break;
                case KEY_NPAGE: pos += slines;
                                break;
                case KEY_HOME : pos = 0;
                                break;
                case KEY_END  : pos = endpos;
                                break;
        }
 }

}

/*
 * Return the C language escape char
 */
static char *int_to_escape( int c)
{

 switch( c) {
         case  0: return "'\\0'";
         case  7: return "'\\a'";
         case  8: return "'\\b'";
         case  9: return "'\\t'";
         case 10: return "'\\n'";
         case 11: return "'\\v'";
         case 12: return "'\\f'";
         case 13: return "'\\r'";
         case 92: return "'\\\\'";
         default: return "    ";
 }

}

/*
 * Convert from integer to binary
 */
static char *int_to_binary( int c)
{
 static char bbuf[ 16];
 register char *bp = bbuf + 8;

 do {
     *--bp = (c % 2) ? '1' : '0';
 } while( (c /= 2) != 0);

 while( bp != bbuf)
        *--bp = '0';

 return bbuf;
}

