/*
   Initialization functions
   --------------------------------------------------------------------
   VCHE RAW - Virtual Console Hex Editor Run Anywhere version

   Copyright (C) 1998, 1999 Diego Javier Grigna <diego@grigna.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

/*
 * Function prototypes
 */
static void init_screen_size( void);
static void init_check_config_file( void);
static void init_process_config( char *buf);
static int init_get_token( void);

extern char *hbuffer;
extern char *hex_buffer;

static char *token;
static const char delimiters[] = "=\t\n";
char *restore_buffer;  /* Save/restore screen data */
int stop_restore;      /* Do not save restore data */

void init( void)
{
 fpos = 0;

 stop_restore = 0;

 init_screen_size();

 signals_init();

 term_init();

 term_hide_cursor();

 init_buffers();

 flag_bell     = 0;
 flag_edit     = 0;
 flag_nohex    = 0;
 flag_readonly = 0;

 init_check_config_file();

 if( flag_readonly && flag_edit) {
     fprintf( stderr, "\nConfig file error: Flags \"readonly\" and \"edit\" cannot be set to \"on\"\nat the same time.\n\n");
     do_exit( 1);
 }

 raw_fill_screen( VC_COLOR_BACKBLUE);

 draw_main_screen();

 set_flag( FLAG_BELL);
 set_flag( FLAG_EDIT);
 set_flag( FLAG_NOHEX);
 set_flag( FLAG_READONLY);

 edit_mask = 0;

 bzero( &srch, 63);

 set_hexmode_variables();
 
 raw_fill_obuffer();
}

static void init_check_config_file( void)
{
 char buf[ 8192];
 FILE *fp;

 /* If the config file does not exist return silently */
 if( access( VCHE_CONFIG_FILE, F_OK) == -1) return;

 if( (fp = fopen( VCHE_CONFIG_FILE, "r")) == NULL) {
     fprintf( stderr, "Can't open \"%s\" : %s\n", VCHE_CONFIG_FILE, strerror( errno));
     do_exit( 1);
 }

 while( !feof( fp) && fgets( buf, sizeof( buf) - 1, fp) != NULL) {
        if( buf[ 0] == '#' || buf[ 0] == '\n') continue;
        init_process_config( buf);
 }

 fclose( fp);
}

static void init_process_config( char *buf)
{
 token = strtok( buf, delimiters);

 if( token == NULL) return;

 if( !strcasecmp( token, "bell"    )) flag_bell     = init_get_token();
 if( !strcasecmp( token, "edit"    )) flag_edit     = init_get_token();
 if( !strcasecmp( token, "nohex"   )) flag_nohex    = init_get_token();
 if( !strcasecmp( token, "readonly")) flag_readonly = init_get_token();
}

static int init_get_token( void)
{
 token = strtok( NULL, delimiters);

 if( token == NULL) return 0;

 if( !strcasecmp( token,  "on")) return 1;

 if( !strcasecmp( token, "off")) return 0;

 return 0;
}

void init_buffers( void)
{
 hex_buffer     = ( char *) allocate_mem( scrn.cols);
 hbuffer        = ( char *) allocate_mem( scrn.cols *     ( scrn.lines - 2));
 obuffer        = ( char *) allocate_mem( scrn.cols * 2 * ( scrn.lines - 2));
 restore_buffer = ( char *) allocate_mem( scrn.cols * 2 *   scrn.lines     );
}

static void init_screen_size( void)
{
 struct winsize wsize;

 /* Defaults, if something goes wrong */
 wsize.ws_row    = 25;
 wsize.ws_col    = 80;
 wsize.ws_xpixel =  0;
 wsize.ws_ypixel =  0;

 /* If we are on a terminal, get windows size */
 if( isatty( fileno( stdin)))
     ioctl( fileno( stdin), TIOCGWINSZ, &wsize);
 
 scrn.lines = wsize.ws_row;
 scrn.cols  = wsize.ws_col;
 scrn.xp    = wsize.ws_xpixel;
 scrn.yp    = wsize.ws_xpixel;
}

