/* 
   Hex viewing & editing functions
   --------------------------------------------------------------------
   VCHE - Virtual Console Hex Editor

   Copyright (C) 1998, 1999 Diego Javier Grigna <diego@grigna.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

#define HEX_LINE_1 "0000000000000000111111111111111122222222222222223333333333333333444444444444444455555555555555556666666666666666777777777777777788888888888888889999999999999999AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCCDDDDDDDDDDDDDDDDEEEEEEEEEEEEEEEEFFFFFFFFFFFFFFFF"
#define HEX_LINE_2 "0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF"

/* It's the fasted way to translate to base 16 */
#if defined(__LITTLE_ENDIAN_BITFIELD)
    char hex_trans_byte1[] = HEX_LINE_1;
    char hex_trans_byte2[] = HEX_LINE_2;
#elif defined(__BIG_ENDIAN_BITFIELD)
    char hex_trans_byte1[] = HEX_LINE_2;
    char hex_trans_byte2[] = HEX_LINE_1;
#else
    char hex_trans_byte1[] = HEX_LINE_1;
    char hex_trans_byte2[] = HEX_LINE_2;
#warning "Please fix <asm/byteorder.h> !!!"
#warning "or mail me the problem <diego@grigna.com> !!!"
#endif

/*
 * Convertion table to base 16
 */
char hex_trans[] = "0123456789ABCDEF";

/* The hex buffer */
char *hbuffer;

/* The output buffer */
char *obuffer;

char *hex_buffer;

long fpos;                             /* File position                    */
int bread;                             /* bytes read                       */

int bytes_to_read;                     /* Quantity of bytes to read        */

int bytes_per_row;                     /* Quantity of ascii bytes per row  */
int bytes_per_row_hex;                 /* Quantity of hex bytes per row    */

void hex_mainloop( int view_or_edit)
{
 int tmp_flag_x = 0;
 int tmp_flag_m = 0;
 int tmp_flag_z = 0;
 char c;

 set_hexmode_variables();

 memset( hex_buffer, ' ', scrn.cols - 1);

 file_open( ll_file_current->filename);
 
 draw_filename( ll_file_current->filename);

 hex_showfile();
 vc_flush();

 term_raw();

 if( view_or_edit) 
     goto editkey;

 while( 1) {
        c = getc( stdin);

        switch( c) {
                /* Quit */
                case 'q': do_exit( 0);
                          break; /* not reached */
                /* Help */
                case '?':
                case 'h':
                case 'H': help();
                          break;
                case 'x': goto nohexmode;
                          break;
                /* Toggle read/write mode */
                case 'w': if( check_readonly() == -1)
                              break;
                          flag_readonly = flag_readonly ? 0 : 1;
                          set_flag( FLAG_READONLY);
                          ll_file_current->filepos = fpos;
                          file_open( ll_file_current->filename);
                          fpos = lseek( fd, (long) ll_file_current->filepos, SEEK_SET);
                          draw_filename( ll_file_current->filename);
                          hex_showfile();
                          vc_flush();
                          break;
                /* Toggle bell */
                case 'b': flag_bell = flag_bell ? 0 : 1;
                          set_flag( FLAG_BELL);
                          break;
                /* Go to next file if any */
                case 'n': ll_file_current->filepos = fpos;
                          if( file_go_next() == -1)
                              break;
                          fpos = lseek( fd, (long) ll_file_current->filepos, SEEK_SET);
                          draw_filename( ll_file_current->filename);
                          hex_showfile();
                          vc_flush();
                          break;
                /* Go to previous file if any */
                case 'p': ll_file_current->filepos = fpos;
                          if( file_go_prev() == -1)
                              break;
                          fpos = lseek( fd, (long) ll_file_current->filepos, SEEK_SET);
                          draw_filename( ll_file_current->filename);
                          hex_showfile();
                          vc_flush();
                          break;
                /* Edit */
                case 'e': goto editkey;
                          break;
                /* Search a string */
                case 's': search();
                          hex_showfile();
                          search_set_color();
                          vc_flush();
                          break;
                /* Search again */
                case ' ': search_again();
                          hex_showfile();
                          search_set_color();
                          vc_flush();
                          break;
                /* Refresh file size */
                case 'r': fstat( fd, &stbuf);
                          break;
                /* Toggle non-printable ascii mask */
                case 'm': flag_mask = flag_mask ? 0 : 1;
                          set_flag( FLAG_MASK);
                          hex_showfile();
                          vc_flush();
                          break;
                /* Toggle zero mask */
                case '0': flag_zero = flag_zero ? 0 : 1;
                          set_flag( FLAG_ZERO);
                          hex_showfile();
                          vc_flush();
                          break;
                /* Goto an offset */
                case 'g': filegoto();
                          hex_showfile();
                          vc_flush();
                          break;
                /* ESC char, process F* and ARROW keys */
                case 27: switch( key_process()) {
                                 /* Help */
                                 case VC_KEY_F1   : help();
                                                    break;
                                 /* Edit */
                                 case VC_KEY_F2   : 
editkey:                                            
                                                    if( flag_readonly) {
                                                        vc_show_alert_box( "We are in READ ONLY mode!", 0);
                                                        break;
                                                    }

                                                    vc_fill_obuffer();

                                                    tmp_flag_m = flag_mask;
                                                    tmp_flag_z = flag_zero;
                                                    tmp_flag_x = flag_nohex;
                                                    flag_edit  = 1;
                                                    flag_mask  = 0;
                                                    flag_zero  = 0;
                                                    flag_nohex = 0;
                                                    set_flag( FLAG_EDIT);
                                                    set_flag( FLAG_MASK);
                                                    set_flag( FLAG_ZERO);
                                                    set_flag( FLAG_NOHEX);
                                                    set_hexmode_variables();
                                                    edit();
                                                    flag_edit  = 0;
                                                    flag_mask  = tmp_flag_m;
                                                    flag_zero  = tmp_flag_z;
                                                    flag_nohex = tmp_flag_x;
                                                    set_flag( FLAG_EDIT);
                                                    set_flag( FLAG_MASK);
                                                    set_flag( FLAG_ZERO);
                                                    set_flag( FLAG_NOHEX);
                                                    set_hexmode_variables();
                                                    tmp_flag_x = tmp_flag_m = tmp_flag_z = 0;
                                                    draw_fkeys();
                                                    hex_showfile();
                                                    vc_flush();
                                                    break;
                                 /* No-hex/hex toggle */
                                 case VC_KEY_F3   : 
nohexmode:
                                                    flag_nohex = flag_nohex ? 0 : 1;
                                                    set_flag( FLAG_NOHEX);
                                                    set_hexmode_variables();
                                                    draw_fkeys();
                                                    vc_fill_obuffer();
                                                    hex_showfile();
                                                    vc_flush();
                                                    break;
                                 /* Search a string */
                                 case VC_KEY_F4   : search();
                                                    hex_showfile();
                                                    search_set_color();
                                                    vc_flush();
                                                    break;
                                 /* Search againg */
                                 case VC_KEY_F5   : search_again();
                                                    hex_showfile();
                                                    search_set_color();
                                                    vc_flush();
                                                    break;
                                 /* Goto an offset */
                                 case VC_KEY_F6   : filegoto();
                                                    hex_showfile();
                                                    vc_flush();
                                                    break;
                                 /* Select file to edit */
                                 case VC_KEY_F7   : ll_file_browse();
                                                    break;
                                 /* Show ASCII table */
                                 case VC_KEY_F8   : ascii_table();
                                                    break;
                                 /* Show GNU GPL */
                                 case VC_KEY_F9   : show_gnu_license();
                                                    break;
                                 /* Quit */
                                 case VC_KEY_F10  : do_exit( 0); 
                                                    break; /* not reached */
                                 case VC_KEY_UP   : file_prev_line();
                                                    vc_flush();
                                                    break;
                                 case VC_KEY_DOWN : file_next_line();
                                                    vc_flush();
                                                    break;
                                 case VC_KEY_RIGHT: file_next_byte();
                                                    vc_flush();
                                                    break;
                                 case VC_KEY_LEFT : file_prev_byte();
                                                    vc_flush();
                                                    break;
                                 case VC_KEY_HOME : file_go_home(); 
                                                    vc_flush();
                                                    break;
                                 case VC_KEY_END  : file_go_end();
                                                    vc_flush();
                                                    break;
                                 case VC_KEY_PGUP : file_prev_block();
                                                    vc_flush();
                                                    break;
                                 case VC_KEY_PGDN : file_next_block();
                                                    vc_flush();
                                                    break;
                                 default: break;
                         }
                         break;
        } /* end switch( c) */

 } /* end while( 1) */

}

void hex_showfile( void)
{
 char offset_buffer[ 32];
 char *ascii_buffer;
 char *zero_buffer;
 long lfpos;
 int startxy;
 int size;
 int line;
 int pos;

 ascii_buffer = ( char *) allocate_mem( scrn.cols);
 zero_buffer  = ( char *) allocate_mem( scrn.cols);

 startxy = scrn.cols - bytes_per_row;

 if( stbuf.st_size > 0)
     draw_offset_n_perc( fpos, stbuf.st_size);

 lfpos = fpos;

 bread = read( fd, hbuffer, bytes_to_read);

 lseek( fd, (long) fpos, SEEK_SET);

 if( bread < bytes_to_read)
     bzero( hbuffer + bread, bytes_to_read - bread);

 for( pos = line = 0; pos < bread; pos += bytes_per_row, line++) {

      hex_fill_offset_buffer( offset_buffer, lfpos);

      vc_color_buffer_puts( 0, line, 10, offset_buffer, VC_COLOR_BACKBLUE);

      size = (bread == bytes_to_read) ? bytes_per_row : ((bread - pos) > bytes_per_row) ? bytes_per_row : bread - pos;

      if( !flag_nohex) {
          hex_fill_hex_buffer( hbuffer + pos, hex_buffer, size);

          /* I need to write colors here, because of the edit routines */
          vc_color_buffer_puts( 12, line, bytes_per_row_hex, hex_buffer, VC_COLOR_BACKBLUE);
      }

      if( flag_mask) {
          hex_fill_ascii_buffer( hbuffer + pos, ascii_buffer, size);
          vc_color_buffer_puts( startxy, line, bytes_per_row, ascii_buffer, VC_COLOR_BACKBLUE);
      } else if( flag_zero) {
                 hex_fill_zero_buffer( hbuffer + pos, zero_buffer, size);
                 vc_color_buffer_puts( startxy, line, bytes_per_row, zero_buffer  , VC_COLOR_BACKBLUE);
             } else
                 vc_color_buffer_puts( startxy, line, bytes_per_row, hbuffer + pos, VC_COLOR_BACKBLUE);

      lfpos += bytes_per_row;
 }

 while( line < scrn.lines - 2)
        vc_color_buffer_puts( 0, line++, scrn.cols, replicate( ' ', scrn.cols), VC_COLOR_BACKBLUE);

 free( ascii_buffer);
 free( zero_buffer );
}

void hex_fill_hex_buffer( char *buf, char *hex_buffer, int size)
{
 int j = 0, i = 0, sp = 1;

 while( j < size) {
        hex_buffer[ i++] = hex_trans_byte1[ (unsigned char) buf[ j  ]];
        hex_buffer[ i++] = hex_trans_byte2[ (unsigned char) buf[ j++]];
        if( sp == 4) {
             i += 2;
             sp = 1;
        } else {
             i++;
             sp++;
        }
 }

 while( j++ < bytes_per_row) {
        hex_buffer[ i++] = ' ';
        hex_buffer[ i++] = ' ';
        if( sp == 4) {
             i += 2;
             sp = 1;
        } else {
             i++;
             sp++;
        }
 }

}

void hex_fill_offset_buffer( char *offset_buffer, int pos)
{
 register char *bp = offset_buffer + 10;

 /* I won't call sprintf, it is a slow function. */
 do {
     *--bp = hex_trans[ pos % 16];
 } while( (pos /= 16) != 0);

 /* pad with zeros */
 while( bp != offset_buffer)
        *--bp = '0';
}

void hex_fill_ascii_buffer( char *buf, char *ascii_buffer, int size)
{
 int i;

 for( i = 0; i < size; i++) 
      ascii_buffer[ i] = buf[ i] > 31 && buf[ i] < 127 ? buf[ i] : '.';

 while( i < bytes_per_row)
        ascii_buffer[ i++] = ' ';
}

void hex_fill_zero_buffer( char *buf, char *zero_buffer, int size)
{
 int i;

 for( i = 0; i < size; i++)
      zero_buffer[ i] = buf[ i] == 0 ? '.': buf[ i];

 while( i < bytes_per_row)
        zero_buffer[ i++] = ' ';
}

