/*
   Miscellaneous functions
   --------------------------------------------------------------------
   VCHE - Virtual Console Hex Editor

   Copyright (C) 1998, 1999 Diego Javier Grigna <diego@grigna.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

static void print_long( int value, int prec);

int flag_bell;     /* Bell flag, set bell on and off               */
int flag_edit;     /* Edit flag                                    */
int flag_mask;     /* Mask flag, mask non-printable characters     */
int flag_zero;     /* Zero flag, mask zero with "."                */
int flag_nohex;    /* No hex flag, toggle viewing the hex segment  */
int flag_readonly; /* Read Only flag, disable/enable readonly mode */

static char trans_buffer[ 32];

void do_exit( int status)
{
 term_color( TERM_COLOR_OFF);
 term_unhide_cursor();
 term_noraw();
 exit( status);
}

void draw_filename( char *filename)
{
 char *ptr;
 int len;
 int mlen;

 ptr = rindex( filename, '/');

 if( ptr == NULL)
     ptr = filename;
 else
     ptr++;

 len  = strlen( ptr);
 mlen = scrn.cols - 26;

 vc_color_puts( 0, 0, mlen, replicate( ' ', mlen), VC_COLOR_BACKCYAN);

 if( len > mlen) {
     vc_color_putc( mlen - 1, 0, '\x1A', VC_COLOR_ATTENTION);
     len = mlen - 1;
 }

 vc_puts( 0, 0, len, ptr);
}

void draw_main_screen( void)
{
 vc_fill_screen( VC_COLOR_BACKBLUE);

 vc_color_puts( 0, 0, scrn.cols, replicate( ' ', scrn.cols), VC_COLOR_BACKCYAN);

 vc_puts( scrn.cols - 26, 0, 8, "WXMEB0");

 vc_putc( scrn.cols - 5 , 0, '');

 vc_putc( scrn.cols - 1 , 0, '%');

 draw_fkeys();
}

void draw_fkeys( void)
{
 int sp, x, y;

 /*
  * I have done it for 132x60 and others non 80xANY modes
  */

 sp = ( scrn.cols / 10) - 8;
 y = scrn.lines - 1;

 vc_color_puts( 0, scrn.lines - 1, scrn.cols, replicate( ' ', scrn.cols), VC_COLOR_BACKCYAN);

 x  = 2;      vc_puts( x, y, 4, "HELP"  ); /* F1  */
 x += 8 + sp; vc_puts( x, y, 4, "EDIT"  ); /* F2  */
 x += 8 + sp;
 if( flag_nohex)                      /* F3  */
     vc_puts( x, y, 5, "HEX  "); 
 else
     vc_puts( x, y, 5, "NOHEX");
 x += 8 + sp; vc_puts( x, y, 6, "SEARCH"); /* F4  */
 x += 8 + sp; vc_puts( x, y, 6, "SAGAIN"); /* F5  */
 x += 8 + sp; vc_puts( x, y, 4, "GOTO"  ); /* F6  */
 x += 8 + sp; vc_puts( x, y, 5, "FILES" ); /* F7  */
 x += 8 + sp; vc_puts( x, y, 6, "ATABLE"); /* F8  */
 x += 8 + sp; vc_puts( x, y, 6, "GNUGPL"); /* F9  */
 x += 8 + sp; vc_puts( x, y, 4, "QUIT"  ); /* F10 */

 x  = 0;      vc_color_puts( x, y, 2, " 1", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, " 2", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, " 3", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, " 4", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, " 5", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, " 6", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, " 7", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, " 8", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, " 9", VC_LIGHTGRAY );
 x += 8 + sp; vc_color_puts( x, y, 2, "10", VC_LIGHTGRAY );
}

void draw_offset_n_perc( long offset, long filesize)
{
 print_long( offset * 100 / filesize, 3);
 vc_puts( scrn.cols -  4, 0,  3, trans_buffer);

 print_long( offset, 13);
 vc_puts( scrn.cols - 18, 0, 13, trans_buffer);
}

void print_long( int value, int prec)
{
 register char *bp = trans_buffer + prec;

 /* Yes I have knowledge of the existence of the printf family of
    functions, but I don't want to call them! they are slow, and
    this part of the program gets called too much time, I do this
    because I hate slow programs!, please use gprof and time to 
    measure your programs.
 */

 do {
     *--bp = hex_trans[ value % 10];
 } while( (value /= 10) != 0);

 /* space padding */
 while( bp != trans_buffer)
        *--bp = ' ';
}


void set_flag( int flag)
{
 int xsflag = scrn.cols - 25;

 switch( flag) {
         case FLAG_ZERO: if( flag_zero)
                             vc_color_putc( xsflag + 5, 0, '0' , VC_COLOR_ATTENTION);
                         else
                             vc_color_putc( xsflag + 5, 0, '0' , VC_COLOR_BACKCYAN);
                         break;
         case FLAG_BELL: if( flag_bell)
                             vc_color_putc( xsflag + 4, 0, 'B' , VC_COLOR_ATTENTION);
                         else
                             vc_color_putc( xsflag + 4, 0, 'B' , VC_COLOR_BACKCYAN);
                         break;
         case FLAG_EDIT: if( flag_edit)
                             vc_color_putc( xsflag + 3, 0, 'E' , VC_COLOR_ATTENTION);
                         else
                             vc_color_putc( xsflag + 3, 0, 'E' , VC_COLOR_BACKCYAN);
                         break;
         case FLAG_MASK: if( flag_mask)
                             vc_color_putc( xsflag + 2, 0, 'M' , VC_COLOR_ATTENTION);
                         else
                             vc_color_putc( xsflag + 2, 0, 'M' , VC_COLOR_BACKCYAN);
                         break;
         case FLAG_NOHEX:
                         if( flag_nohex)
                             vc_color_putc( xsflag + 1, 0, 'X' , VC_COLOR_ATTENTION);
                         else
                             vc_color_putc( xsflag + 1, 0, 'X' , VC_COLOR_BACKCYAN);
                         break;
         case FLAG_READONLY:
                         if( flag_readonly)
                             vc_color_putc( xsflag, 0, 'W' , VC_COLOR_ATTENTION);
                         else
                             vc_color_putc( xsflag, 0, 'W' , VC_COLOR_BACKCYAN);
                         break;
 }
}

void *allocate_mem( int size)
{
 void *voidptr = NULL;

 voidptr = calloc( 1, size + 1);

 if( voidptr == NULL) {
     fprintf( stderr, "Not enough memory\n");
     do_exit( 1);
 }

 return voidptr;
}

char *trim_spaces( char *s)
{
 static char a[ 8192];
 char *save = s;
 char *start, *end;

 start = s;
 end = strchr( s, 0);

 if( *end == '\n' || *end == '\r') *end = 0;

 while( end >= start && *end <= ' ') end--;
 *++end = 0;

 while( end >= start && *start <= ' ') start++;
 save = start;

 bzero( a, sizeof( a));
 strcpy( a, save);

 return a;
}

void set_hexmode_variables( void)
{

 bytes_per_row = scrn.cols - 12;

 if( !flag_nohex) {
     bytes_per_row     = bytes_per_row * 4 / 17;
     bytes_per_row_hex = bytes_per_row * 13 / 4;
 }

 bytes_to_read = bytes_per_row * (scrn.lines - 2);
}

int check_readonly( void)
{
 int i;
 if( !flag_readonly) return 1;

 term_beep();
 i = vc_ask( "We are in READ ONLY mode do you want to set READ/WRITE mode?", " Yes ", " No ", " Cancel ", 1);

 term_hide_cursor();

 if( i == 0) return 1;

 return -1;
}

char *replicate( char c, int size)
{
 static char rstr[ 2048];

 memset( &rstr, c, size);

 return rstr;
}

